"""This module provides microscope image stitching with the algorithm by MIST."""
import warnings
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from ._constrained_refinement import refine_translations
from ._global_optimization import compute_final_position
from ._global_optimization import compute_maximum_spanning_tree
from ._stage_model import compute_image_overlap
from ._stage_model import compute_repeatability
from ._stage_model import filter_by_repeatability
from ._stage_model import replace_invalid_translations
from ._translation_computation import interpret_translation
from ._translation_computation import multi_peak_max
from ._translation_computation import pcm
from ._typing_utils import Float
from ._typing_utils import NumArray


def stitch_images(
    images: Union[Sequence[NumArray], NumArray],
    rows: Optional[Sequence[Any]] = None,
    cols: Optional[Sequence[Any]] = None,
    position_indices: Optional[NumArray] = None,
    pou: Float = 3,
    overlap_prob_uniform_threshold: Float = 80,
    full_output: bool = False,
    row_col_transpose: bool = True,
) -> Tuple[pd.DataFrame, dict]:
    """Compute image positions for stitching.

    Parameters
    ---------
    images : np.ndarray
        the images to stitch.

    rows : list, optional
        the row indices (tile position in the last dimension) of the images.

    cols : list, optional
        the column indices (tile position in the second last dimension) of the images

    position_indices : np.ndarray, optional
        the tile position indices in each dimension.
        the dimensions corresponds to (image, index)
        ignored if rows and cols are not None.

    pou : Float, default 3
        the "percent overlap uncertainty" parameter

    overlap_prob_uniform_threshold : Float, default 80
        the upper threshold for "uniform probability".
        raise this value to ease assumption
        that the displacement is following non-uniform distribution.

    full_output : bool, default False
        if True, returns the full comptutation result in the pd.DataFrame

    row_col_transpose : bool, default True
        if True, row and col indices are switched.
        only for compatibility and the default value will be False in the future.

    Returns
    -------
    grid : pd.DataFrame
        the result dataframe with the rows "x_pos" and "y_pos" whose values are
        the absolute positions.

    prop_dict : dict
        the dict of estimated parameters. (to be documented)
    """
    images = np.array(images)
    assert (position_indices is not None) or (rows is not None and cols is not None)
    if position_indices is None:
        if row_col_transpose:
            warnings.warn(
                "row_col_transpose is True. The default value will be changed to False in the major release."
            )
            position_indices = np.array([rows, cols]).T
        else:
            position_indices = np.array([cols, rows]).T
    position_indices = np.array(position_indices)
    assert images.shape[0] == position_indices.shape[0]
    assert position_indices.shape[1] == images.ndim - 1
    assert 0 <= overlap_prob_uniform_threshold and overlap_prob_uniform_threshold <= 100
    _rows, _cols = position_indices.T

    W, H = images.shape[1:]

    grid = pd.DataFrame(
        {
            "row": _rows,
            "col": _cols,
        },
        index=np.arange(len(_rows)),
    )

    def get_index(row, col):
        df = grid[(grid["row"] == row) & (grid["col"] == col)]
        assert len(df) < 2
        if len(df) == 1:
            return df.index[0]
        else:
            return None

    grid["top"] = grid.apply(
        lambda g: get_index(g["row"] - 1, g["col"]), axis=1
    ).astype(pd.Int32Dtype())
    grid["left"] = grid.apply(
        lambda g: get_index(g["row"], g["col"] - 1), axis=1
    ).astype(pd.Int32Dtype())

    ###### translationComputation ######
    for direction in ["top", "left"]:
        for i2, g in tqdm(grid.iterrows(), total=len(grid)):
            i1 = g[direction]
            if pd.isna(i1):
                continue
            image1 = images[i1]
            image2 = images[i2]

            PCM = pcm(image1, image2).real
            found_peaks = list(zip(*multi_peak_max(PCM)))

            interpreted_peaks = []
            for r, c, _ in found_peaks:
                interpreted_peaks.append(interpret_translation(image1, image2, r, c))
            max_peak = interpreted_peaks[np.argmax(np.array(interpreted_peaks)[:, 0])]
            for j, key in enumerate(["ncc", "x", "y"]):
                grid.loc[i2, f"{direction}_{key}_first"] = max_peak[j]

    prob_uniform_n, mu_n, sigma_n = compute_image_overlap(
        grid, "top", W, H, prob_uniform_threshold=overlap_prob_uniform_threshold
    )
    overlap_n = 100 - mu_n
    prob_uniform_w, mu_w, sigma_w = compute_image_overlap(
        grid, "left", W, H, prob_uniform_threshold=overlap_prob_uniform_threshold
    )
    overlap_w = 100 - mu_w

    overlap_n = np.clip(overlap_n, pou, 100 - pou)
    overlap_w = np.clip(overlap_w, pou, 100 - pou)

    grid, r = compute_repeatability(grid, overlap_n, overlap_w, W, H, pou)
    grid = filter_by_repeatability(grid, r)
    grid = replace_invalid_translations(grid)

    grid = refine_translations(images, grid, r)

    tree = compute_maximum_spanning_tree(grid)
    grid = compute_final_position(grid, tree)

    prop_dict = {
        "W": W,
        "H": H,
        "overlap_top": overlap_n,
        "overlap_left": overlap_w,
        "overlap_top_results": {
            "prob_uniform": prob_uniform_n,
            "mu": mu_n,
            "sigma": sigma_n,
        },
        "overlap_left_results": {
            "prob_uniform": prob_uniform_w,
            "mu": mu_w,
            "sigma": sigma_w,
        },
        "repeatability": r,
    }
    if full_output:
        return grid, prop_dict
    else:
        return grid[["col", "row", "y_pos", "x_pos"]], prop_dict



import itertools
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from ._typing_utils import Float
from ._typing_utils import FloatArray
from ._typing_utils import Int


def calc_liklihood(prob_uniform: Float, mu: Float, sigma: Float, t: Float) -> Float:
    """Calculate the liklihood of the translation.

    Parameters
    ----------
    prob_uniform : Float
        the probability of the uniform error
    mu : Float
        the mean of the Gaussian distribution
    sigma : Float
        the stdev of the Gaussian distribution
    t : Float
        the translation amplitude

    Returns
    -------
    liklihood : Float
        the calculated liklihood
    """
    t2 = -((t - mu) ** 2) / (2 * sigma ** 2)
    norm_liklihood = 1.0 / (np.sqrt(2 * np.pi) * sigma) * np.exp(t2)
    uniform_liklihood = 1 / 100.0
    p = prob_uniform / 100.0
    return p * uniform_liklihood + (1 - p) * norm_liklihood


def compute_inv_liklihood(
    params: Union[Tuple[Float, Float, Float], FloatArray], T: FloatArray
) -> Float:
    """Compute the negated sum of log liklihood.

    Parameters
    ----------
    params : Union[Tuple[Float, Float, Float], FloatArray]
        the parameters : prob_uniform, mu, sigma
    T : FloatArray
        the value of translations

    Returns
    -------
    inv_log_likelihood : Float
        the negated sum of log liklihood
    """
    prob_uniform, mu, sigma = params
    return -np.sum(
        [np.log(np.abs(calc_liklihood(prob_uniform, mu, sigma, t))) for t in T]
    )


def compute_image_overlap(
    grid: pd.DataFrame,
    direction: str,
    W: Int,
    H: Int,
    max_stall_count: Int = 100,
    prob_uniform_threshold: Float = 80,
) -> Tuple[float, ...]:
    """Compute the value of the image overlap.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position, with columns "{top|left}_{x|y}_first"
    direction : str
        the direction of the overlap, either of "top" or "left"
    W : Int
        the image width
    H : Int
        the image height
    max_stall_count : Int, optional
        the maximum count of optimization, by default 100
    prob_uniform_threshold : Float, optional
        the threshold for the probabillity for uniform error, by default 80

    Returns
    -------
    x : Tuple[float, float, float]
        the computed parameters : prob_uniform, mu, sigma

    Raises
    ------
    ValueError
        when direction is not in ["top","left"], raises ValueError
    """
    if direction == "top":
        T = grid["top_y_first"].values / H * 100
    elif direction == "left":
        T = grid["left_x_first"].values / W * 100
    else:
        raise ValueError("direction must be either of top or left")
    T = T[np.isfinite(T)]
    #    print(T)
    #    assert np.all(0 <= T < 100)

    best_model = None
    for _ in range(max_stall_count):
        init_guess = np.random.uniform(0, 100, size=(3,))
        model = minimize(
            compute_inv_liklihood,
            init_guess,
            args=(T,),
            bounds=[(0, 100)] * 3,
            method="trust-constr",
        )
        if model["x"][0] < prob_uniform_threshold:
            if best_model is None:
                best_model = model
            else:
                if np.isclose(best_model["fun"], model["fun"]):
                    assert len(model["x"]) == 3
                    return tuple(map(float, model["x"]))
                elif model["fun"] < best_model["fun"]:
                    best_model = model
    assert (
        not best_model is None
    ), "Overlap model estimation failed, try raising the value of overlap_prob_uniform_threshold"  # noqa :
    #    best_model_params : Tuple[Float,Float,Float] = tuple(best_model["x"])
    assert len(best_model["x"]) == 3
    return tuple(map(float, best_model["x"]))


def filter_by_overlap_and_correlation(
    T: pd.Series, ncc: pd.Series, overlap: Float, size: Int, pou: Float = 3
) -> pd.Series:
    """Filter the translation values by estimated overlap.

    Parameters
    ----------
    T : pd.Series
        the translation
    ncc : pd.Series
        the normalized cross correlation, this value should be > 0.5 to be valid
    overlap : Float
        the estimated overlap
    size : Int
        the size of image dimension
    pou : Float, optional
        the percentile margin for error, by default 3

    Returns
    -------
    isvalid : pd.Series
        whether the translation is within the estimated limit
    """
    r = (size * (100 - overlap - pou) / 100, size * (100 - overlap + pou) / 100)
    return (T.between(*r)) & (ncc > 0.5)


def filter_outliers(T: pd.Series, isvalid: pd.Series, w: Float = 1.5) -> pd.Series:
    """Filter the translation outside the 25% and 75% percentiles * w.

    Parameters
    ----------
    T : pd.Series
        the translation
    isvalid : pd.Series
        whether the translation is valid
    w : Float, optional
        the coef for the percentiles, by default 1.5

    Returns
    -------
    isvalid : pd.Series
        whether the translation is within the estimated limit
    """
    valid_T = T[isvalid].values
    if len(valid_T) < 1:
        return isvalid
    q1, _, q3 = np.quantile(valid_T, (0.25, 0.5, 0.75))
    iqd = max(1, np.abs(q3 - q1))
    return isvalid & T.between(q1 - w * iqd, q3 + w * iqd)


def compute_repeatability(
    grid: pd.DataFrame, overlap_n: Float, overlap_w: Float, W: Int, H: Int, pou: Float
) -> Tuple[pd.DataFrame, Float]:
    """Compute the repeatability of the stage motion.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position, with columns "{top|left}_{x|y|ncc}_first"
    overlap_n : Float
        the estimated overlap for top direction
    overlap_w : Float
        the estimated overlap for left direction
    W : Int
        the width of the height images
    H : Int
        the width of the tile images
    pou : Float
        the percentile margin for error, by default 3

    Returns
    -------
    grid : pd.DataFrame
        the updated dataframe for the grid position
    repeatability : Float
        the repeatability of the stage motion
    """
    grid["top_valid1"] = filter_by_overlap_and_correlation(
        grid["top_y_first"], grid["top_ncc_first"], overlap_n, H, pou
    )
    grid["left_valid1"] = filter_by_overlap_and_correlation(
        grid["left_x_first"], grid["left_ncc_first"], overlap_w, W, pou
    )
    grid["top_valid2"] = filter_outliers(grid["top_y_first"], grid["top_valid1"])
    grid["left_valid2"] = filter_outliers(grid["left_x_first"], grid["left_valid1"])

    if np.any(grid["top_valid2"]):
        xs = grid[grid["top_valid2"]]["top_x_first"]
        rx_top = np.ceil((xs.max() - xs.min()) / 2)
        _, yss = zip(*grid[grid["top_valid2"]].groupby("col")["top_y_first"])
        ry_top = np.ceil(np.max([np.max(ys) - np.min(ys) for ys in yss]) / 2)
        r_top = max(rx_top, ry_top)
    else:
        r_top = 0  # better than failing

    if np.any(grid["left_valid2"]):
        ys = grid[grid["left_valid2"]]["left_y_first"]
        ry_left = np.ceil((ys.max() - ys.min()) / 2)
        _, xss = zip(*grid[grid["left_valid2"]].groupby("row")["left_x_first"])
        rx_left = np.ceil(np.max([np.max(xs) - np.min(xs) for xs in xss]) / 2)
        r_left = max(ry_left, rx_left)
    else:
        r_left = 0  # better than failing

    return grid, max(r_top, r_left)


def filter_by_repeatability(grid: pd.DataFrame, r: Float) -> pd.DataFrame:
    """Filter the stage translation by repeatability.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position, with columns "{top|left}_{x|y|ncc}_first"
    r : Float
        the repeatability value

    Returns
    -------
    grid : pd.DataFrame
        the updated dataframe for the grid position
    """
    for _, grp in grid.groupby("row"):
        isvalid = grp["top_valid2"].astype(bool)
        if not any(isvalid):
            grid.loc[grp.index, "top_valid3"] = False
        else:
            medx = grp[isvalid]["top_x_first"].median()
            medy = grp[isvalid]["top_y_first"].median()
            grid.loc[grp.index, "top_valid3"] = (
                grp["top_x_first"].between(medx - r, medx + r)
                & grp["top_y_first"].between(medy - r, medy + r)
                & (grp["top_ncc_first"] > 0.5)
            )
    for _, grp in grid.groupby("col"):
        isvalid = grp["left_valid2"]
        if not any(isvalid):
            grid.loc[grp.index, "left_valid3"] = False
        else:
            medx = grp[isvalid]["left_x_first"].median()
            medy = grp[isvalid]["left_y_first"].median()
            grid.loc[grp.index, "left_valid3"] = (
                grp["left_x_first"].between(medx - r, medx + r)
                & grp["left_y_first"].between(medy - r, medy + r)
                & (grp["left_ncc_first"] > 0.5)
            )
    return grid


def replace_invalid_translations(grid: pd.DataFrame) -> pd.DataFrame:
    """Replace invalid translations by estimated values.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position,
        with columns "{top|left}_{x|y}_second" and "{top|left}_valid3"

    Returns
    -------
    grid : pd.DataFrame
        the updatd dataframe for the grid position
    """
    for direction in ["top", "left"]:
        for key in ["x", "y", "ncc"]:
            isvalid = grid[f"{direction}_valid3"]
            grid.loc[isvalid, f"{direction}_{key}_second"] = grid.loc[
                isvalid, f"{direction}_{key}_first"
            ]
    for direction, rowcol in zip(["top", "left"], ["row", "col"]):
        for _, grp in grid.groupby(rowcol):
            isvalid = grp[f"{direction}_valid3"].astype(bool)
            if any(isvalid):
                assert all(
                    pd.isna(grid.loc[grp.index[~isvalid], f"{direction}_x_second"])
                )
                assert all(
                    pd.isna(grid.loc[grp.index[~isvalid], f"{direction}_y_second"])
                )
                grid.loc[grp.index[~isvalid], f"{direction}_x_second"] = grp[isvalid][
                    f"{direction}_x_first"
                ].median()
                grid.loc[grp.index[~isvalid], f"{direction}_y_second"] = grp[isvalid][
                    f"{direction}_y_first"
                ].median()
                grid.loc[grp.index[~isvalid], f"{direction}_ncc_second"] = -1
    for direction, xy in itertools.product(["top", "left"], ["x", "y"]):
        key = f"{direction}_{xy}_second"
        isna = pd.isna(grid[key])
        grid.loc[isna, key] = grid.loc[~isna, key].median()
        grid.loc[isna, f"{direction}_ncc_second"] = -1
    for direction, xy in itertools.product(["top", "left"], ["x", "y"]):
        assert np.all(np.isfinite(grid[f"{direction}_{xy}_second"]))

    return grid



"""Command-line interface."""
import click

# from .stitching import stitch_images


@click.command()
@click.version_option()
def main() -> None:
    """M2Stitch."""


#    testimages, props = test_image_path
#    """It exits with a status code of zero."""
#    rows = props["row"].to_list()
#    cols = props["col"].to_list()
#    result_df, _ = stitch_images(testimages, rows, cols)


if __name__ == "__main__":
    main(prog_name="m2stitch")  # pragma: no cover



import networkx as nx
import numpy as np
import pandas as pd

from ._typing_utils import Int


def compute_maximum_spanning_tree(grid: pd.DataFrame) -> nx.Graph:
    """Compute the maximum spanning tree for grid position determination.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position,
        with columns "{top|left}_{x|y|ncc|valid3}"

    Returns
    -------
    tree : nx.Graph
        the result spanning tree
    """
    connection_graph = nx.Graph()
    for i, g in grid.iterrows():
        for direction in ["top", "left"]:
            if not pd.isna(g[direction]):
                weight = g[f"{direction}_ncc"]
                if g[f"{direction}_valid3"]:
                    weight = weight + 10
                connection_graph.add_edge(
                    i,
                    g[direction],
                    weight=weight,
                    direction=direction,
                    f=i,
                    t=g[direction],
                    x=g[f"{direction}_x"],
                    y=g[f"{direction}_y"],
                )
    return nx.maximum_spanning_tree(connection_graph)


def compute_final_position(
    grid: pd.DataFrame, tree: nx.Graph, source_index: Int = 0
) -> pd.DataFrame:
    """Compute the final tile positions by the computed maximum spanning tree.

    Parameters
    ----------
    grid : pd.DataFrame
        the dataframe for the grid position
    tree : nx.Graph
        the maximum spanning tree
    source_index : Int, optional
        the source position of the spanning tree, by default 0

    Returns
    -------
    grid : pd.DataFrame
        the result dataframe for the grid position, with columns "{x|y}_pos"
    """
    grid.loc[source_index, "x_pos"] = 0
    grid.loc[source_index, "y_pos"] = 0

    nodes = [source_index]
    walked_nodes = []
    while len(nodes) > 0:
        node = nodes.pop()
        walked_nodes.append(node)
        for adj, props in tree.adj[node].items():
            if not adj in walked_nodes:
                assert (props["f"] == node) & (props["t"] == adj) or (
                    props["t"] == node
                ) & (props["f"] == adj)
                nodes.append(adj)
                x_pos = grid.loc[node, "x_pos"]
                y_pos = grid.loc[node, "y_pos"]

                if node == props["t"]:
                    grid.loc[adj, "x_pos"] = x_pos + props["x"]
                    grid.loc[adj, "y_pos"] = y_pos + props["y"]
                else:
                    grid.loc[adj, "x_pos"] = x_pos - props["x"]
                    grid.loc[adj, "y_pos"] = y_pos - props["y"]
    assert not any(pd.isna(grid["x_pos"]))
    assert not any(pd.isna(grid["y_pos"]))
    grid["x_pos"] = grid["x_pos"] - grid["x_pos"].min()
    grid["y_pos"] = grid["y_pos"] - grid["y_pos"].min()
    grid["x_pos"] = grid["x_pos"].astype(np.int32)
    grid["y_pos"] = grid["y_pos"].astype(np.int32)

    return grid



import itertools
from typing import Callable
from typing import Tuple

import numpy as np
import pandas as pd
from tqdm import tqdm

from ._translation_computation import extract_overlap_subregion
from ._translation_computation import ncc
from ._typing_utils import Float
from ._typing_utils import FloatArray
from ._typing_utils import Int
from ._typing_utils import NumArray


def find_local_max_integer_constrained(
    func: Callable[[FloatArray], Float],
    init_x: FloatArray,
    limits: FloatArray,
    max_iter: Int = 100,
) -> Tuple[FloatArray, Float]:
    """Find local maxima of a function with integer steps.

    Parameters
    ----------
    func : Callable[[FloatArray], Float]
        function to optimize
    init_x : FloatArray
        the initial guess of parameters
    limits : FloatArray
        the limit of parameters
    max_iter : Int, optional
        the maximum iteration for the optimization, by default 100

    Returns
    -------
    x : FloatArray
        the optimized parameters
    val : Float
        the optimized value
    """
    init_x = np.array(init_x)
    limits = np.array(limits)
    dim = init_x.shape[0]
    assert limits.shape[0] == dim
    value = func(init_x)
    x = init_x
    for _ in range(max_iter):
        around_x = [x + np.array(dxs) for dxs in itertools.product(*[[-1, 0, 1]] * dim)]
        around_x = [
            x
            for x in around_x
            if np.all(limits[:, 0] <= x) and np.all(x <= limits[:, 1])
        ]
        around_values = np.array(list(map(func, around_x)))

        max_ind = np.argmax(around_values)
        max_x = around_x[max_ind]
        max_value = around_values[max_ind]
        if max_value <= value:
            return x, value
        else:
            x = max_x
            value = max_value
    return x, value


def refine_translations(images: NumArray, grid: pd.DataFrame, r: Float) -> pd.DataFrame:
    """Refine the translations using the normalized cross correlation.

    Parameters
    ----------
    images : NumArray
        the tile images
    grid : pd.DataFrame
        the dataframe for the grid position, with columns "{top|left}_{x|y}_second"
    r : Float
        the repeatability

    Returns
    -------
    grid : pd.DataFrame
        the refined grid position, with columns "{top|left}_{x|y|ncc}"
    """
    for direction in ["top", "left"]:
        for i2, g in tqdm(grid.iterrows(), total=len(grid)):
            i1 = g[direction]
            if pd.isna(i1):
                continue
            image1 = images[i1]
            image2 = images[i2]
            W, H = image1.shape

            def overlap_ncc(params):
                x, y = params
                subI1 = extract_overlap_subregion(image1, x, y)
                subI2 = extract_overlap_subregion(image2, -x, -y)
                return ncc(subI1, subI2)

            init_values = [
                int(g[f"{direction}_x_second"]),
                int(g[f"{direction}_y_second"]),
            ]
            limits = [
                [max(-W + 1, init_values[0] - r), min(W - 1, init_values[0] + r)],
                [max(-H + 1, init_values[1] - r), min(H - 1, init_values[1] + r)],
            ]
            values, ncc_value = find_local_max_integer_constrained(
                overlap_ncc, np.array(init_values), np.array(limits)
            )
            grid.loc[i2, f"{direction}_x"] = values[0]
            grid.loc[i2, f"{direction}_y"] = values[1]
            grid.loc[i2, f"{direction}_ncc"] = ncc_value
    for direction in ["top", "left"]:
        for xy in ["x", "y"]:
            key = f"{direction}_{xy}"
            grid[key] = grid[key].astype(pd.Int32Dtype())
    return grid



from typing import Any
from typing import Union

import numpy as np
import numpy.typing as npt

NumArray = npt.NDArray[Any]
FloatArray = npt.NDArray[np.float_]
IntArray = npt.NDArray[np.int_]

Int = Union[int, np.int_]
Float = Union[float, np.float_]



import itertools
from typing import Tuple

import numpy as np
import numpy.typing as npt

from ._typing_utils import Float
from ._typing_utils import FloatArray
from ._typing_utils import Int
from ._typing_utils import IntArray
from ._typing_utils import NumArray


def pcm(image1: NumArray, image2: NumArray) -> FloatArray:
    """Compute peak correlation matrix for two images.

    Parameters
    ---------
    image1 : np.ndarray
        the first image (the dimension must be 2)

    image2 : np.ndarray
        the second image (the dimension must be 2)

    Returns
    -------
    PCM : np.ndarray
        the peak correlation matrix
    """
    assert image1.ndim == 2
    assert image2.ndim == 2
    assert np.array_equal(image1.shape, image2.shape)
    F1 = np.fft.fft2(image1)
    F2 = np.fft.fft2(image2)
    FC = F1 * np.conjugate(F2)
    return np.fft.ifft2(FC / np.abs(FC)).real.astype(np.float32)


def multi_peak_max(
    PCM: FloatArray, n: int = 2
) -> Tuple[IntArray, IntArray, FloatArray]:
    """Find the first to n th largest peaks in PCM.

    Parameters
    ---------
    PCM : np.ndarray
        the peak correlation matrix
    n : Int
        the number of the peaks


    Returns
    -------
    rows : np.ndarray
        the row indices for the peaks
    cols : np.ndarray
        the column indices for the peaks
    vals : np.ndarray
        the values of the peaks
    """
    row, col = np.unravel_index(np.argsort(PCM.ravel()), PCM.shape)
    vals: FloatArray = PCM[row[-n:][::-1], col[-n:][::-1]]
    return row[-n:][::-1], col[-n:][::-1], vals


def ncc(image1: NumArray, image2: NumArray) -> Float:
    """Compute the normalized cross correlation for two images.

    Parameters
    ---------
    image1 : np.ndarray
        the first image (the dimension must be 2)

    image2 : np.ndarray
        the second image (the dimension must be 2)

    Returns
    -------
    ncc : Float
        the normalized cross correlation
    """
    assert image1.ndim == 2
    assert image2.ndim == 2
    assert np.array_equal(image1.shape, image2.shape)
    image1 = image1.flatten()
    image2 = image2.flatten()
    n = np.dot(image1 - np.mean(image1), image2 - np.mean(image2))
    d = np.linalg.norm(image1) * np.linalg.norm(image2)
    return n / d


def extract_overlap_subregion(image: NumArray, x: Int, y: Int) -> NumArray:
    """Extract the overlapping subregion of the image.

    Parameters
    ---------
    image : np.ndarray
        the image (the dimension must be 2)
    x : Int
        the x position
    y : Int
        the y position

    Returns
    -------
    subimage : np.ndarray
        the extracted subimage
    """
    W = image.shape[0]
    H = image.shape[1]
    assert (np.abs(x) < W) and (np.abs(y) < H)
    xstart = int(max(0, min(x, W, key=int), key=int))
    xend = int(max(0, min(x + W, W, key=int), key=int))
    ystart = int(max(0, min(y, H, key=int), key=int))
    yend = int(max(0, min(y + H, H, key=int), key=int))
    return image[xstart:xend, ystart:yend]


def interpret_translation(
    image1: NumArray, image2: npt.NDArray, xin: Int, yin: Int
) -> Tuple[float, int, int]:
    """Interpret the translation to find the translation with heighest ncc.

    Parameters
    ---------
    image1 : np.ndarray
        the first image (the dimension must be 2)
    image2 : np.ndarray
        the second image (the dimension must be 2)
    xin : Int
        the x position estimated by PCM
    yin : Int
        the y position estimated by PCM

    Returns
    -------
    _ncc : Float
        the highest ncc
    x : Int
        the selected x position
    y : Int
        the selected y position
    """
    assert image1.ndim == 2
    assert image2.ndim == 2
    assert np.array_equal(image1.shape, image2.shape)
    _ncc = -np.infty
    x = 0
    y = 0
    W = image1.shape[0]
    H = image1.shape[1]
    assert 0 <= xin and xin < W
    assert 0 <= yin and yin < H
    xmags = [xin, W - xin] if xin > 0 else [xin]
    ymags = [yin, H - yin] if yin > 0 else [yin]
    for xmag, ymag, xsign, ysign in itertools.product(xmags, ymags, [-1, +1], [-1, +1]):
        subI1 = extract_overlap_subregion(image1, (xmag * xsign), (ymag * ysign))
        subI2 = extract_overlap_subregion(image2, -(xmag * xsign), -(ymag * ysign))
        ncc_val = ncc(subI1, subI2)
        if ncc_val > _ncc:
            _ncc = float(ncc_val)
            x = int(xmag * xsign)
            y = int(ymag * ysign)
    return _ncc, x, y



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['m2stitch']

package_data = \
{'': ['*']}

install_requires = \
['click>=7,<9',
 'networkx>=2.5.1,<3.0.0',
 'numpy>=1.22,<2.0',
 'pandas-stubs>=1.2.0,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'scipy>=1.6.3,<2.0.0',
 'tqdm>=4.60.0,<5.0.0']

entry_points = \
{'console_scripts': ['m2stitch = m2stitch.__main__:main']}

setup_kwargs = {
    'name': 'm2stitch',
    'version': '0.3.0',
    'description': 'M2Stitch',
    'long_description': "M2Stitch\n========\n\n|PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black| |Zenodo|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/m2stitch.svg\n   :target: https://pypi.org/project/m2stitch/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/m2stitch\n   :target: https://pypi.org/project/m2stitch\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/m2stitch\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/m2stitch/latest.svg?label=Read%20the%20Docs\n   :target: https://m2stitch.readthedocs.io/\n   :alt: Read the documentation at https://m2stitch.readthedocs.io/\n.. |Tests| image:: https://github.com/yfukai/m2stitch/workflows/Tests/badge.svg\n   :target: https://github.com/yfukai/m2stitch/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/yfukai/m2stitch/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/yfukai/m2stitch\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n.. |Zenodo| image:: https://zenodo.org/badge/DOI/10.5281/zenodo.5139597.svg\n   :target: https://doi.org/10.5281/zenodo.5139597\n   :alt: Zenodo\n\n**Note** :memo: : one might also be interested in another Python-written stitching tool\nASHLAR_ (bioRxiv_),\nwith a comparable performance to that of MIST and additional features.\n\n.. _ASHLAR: https://github.com/labsyspharm/ashlar\n.. _bioRxiv: https://www.biorxiv.org/content/10.1101/2021.04.20.440625v1\n\nFeatures\n--------\n\n- Provides robust stitching of tiled microscope images on a regular grid, mostly following algorithm by MIST_.\n- Supports missing tiles.\n\nInstallation\n------------\n\nYou can install *M2Stitch* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install m2stitch\n\n\nUsage\n-----\n\nPlease see the Usage_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*M2Stitch* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis program is an unofficial implementation of MIST_ stitching algorithm on GitHub_. The original paper is here_.\n\nThis project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _MIST: https://pages.nist.gov/MIST\n.. _GitHub: https://github.com/usnistgov/MIST\n.. _here: https://github.com/USNISTGOV/MIST/wiki/assets/mist-algorithm-documentation.pdf\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/yfukai/m2stitch/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://m2stitch.readthedocs.io/en/latest/usage.html\n",
    'author': 'Yohsuke Fukai',
    'author_email': 'ysk@yfukai.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yfukai/m2stitch',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""M2Stitch.

A MIST-inspired unofficial stitching imprementation in Python for
microscope images
"""

__author__ = """Yohsuke T. Fukai"""
__email__ = "ysk@yfukai.net"

from .stitching import stitch_images

__all__ = ["stitch_images"]
